import gzip
import os
import shutil
from os import path as _p
from tempfile import TemporaryDirectory
from typing import List
from unittest import TestCase, SkipTest

from ..mountutil import mount, extract, archive, MountableType, ArchiveType, DiskImage


class TestMountable(TestCase):
    test_dir: TemporaryDirectory
    test_vectors: List[str]

    @classmethod
    def setUpClass(cls):
        if cls is TestMountable:
            raise SkipTest("Skip BaseTest tests, it's a base class")
        super(TestMountable, cls).setUpClass()

    def setUp(self):
        self.test_dir = TemporaryDirectory()
        mountables_dir = _p.join(_p.dirname(__file__), 'mountables')
        for test_vector in self.test_vectors:
            if test_vector in os.listdir(mountables_dir):
                shutil.copyfile(_p.join(mountables_dir, test_vector), _p.join(self.test_dir.name, test_vector))
            elif test_vector + r'.gz' in os.listdir(mountables_dir):
                with gzip.open(_p.join(mountables_dir, test_vector + r'.gz'), 'rb') as f_in, \
                        open(_p.join(self.test_dir.name, test_vector), 'wb') as f_out:
                    shutil.copyfileobj(f_in, f_out)

        # print(sh.ls(self.test_dir.name))

    def usage_with(self, file: str, type: MountableType = MountableType.AUTO, gt: str = 'hello'):
        with mount(file, type=type) as m:
            with open(_p.join(m.name, 'hello'), 'r') as f:
                assert f.read().strip() == gt

    def usage_no_with(self, file: str, type: MountableType = MountableType.AUTO, gt: str = 'hello'):
        m = mount(file, type=type)

        with open(_p.join(m.name, 'hello'), 'r') as f:
            assert f.read().strip() == gt

        m.umount()


class TestArchive(TestMountable):
    test_dir: TemporaryDirectory
    test_vectors: List[str]

    @classmethod
    def setUpClass(cls):
        if cls is TestArchive:
            raise SkipTest("Skip BaseTest tests, it's a base class")
        super(TestArchive, cls).setUpClass()

    def usage_extract(self, file: str, type: ArchiveType = ArchiveType.AUTO, gt: str = 'hello'):
        tmpdir = TemporaryDirectory()
        extract(file, tmpdir.name, type=type)

        with open(_p.join(tmpdir.name, 'hello'), 'r') as f:
            assert f.read().strip() == gt


class TestLinuxDiskImage(TestMountable):
    def __init__(self, *args, **kwargs):
        self.test_vectors = ['ext4.img']
        super().__init__(*args, **kwargs)

    def test_with(self):
        self.usage_with(_p.join(self.test_dir.name, 'ext4.img'), type=MountableType.LINUX_DISK_IMAGE)

    def test_no_with(self):
        self.usage_no_with(_p.join(self.test_dir.name, 'ext4.img'), type=MountableType.LINUX_DISK_IMAGE)

    def test_implicit_type(self):
        self.usage_with(_p.join(self.test_dir.name, 'ext4.img'))


class TestYaffs2Oob(TestArchive):
    def __init__(self, *args, **kwargs):
        self.test_vectors = ['yaffs2_OOB.img']
        super().__init__(*args, **kwargs)

    def test_with(self):
        self.usage_with(_p.join(self.test_dir.name, 'yaffs2_OOB.img'), type=MountableType.YAFFS2_ARCHIVE)

    def test_no_with(self):
        self.usage_no_with(_p.join(self.test_dir.name, 'yaffs2_OOB.img'), type=MountableType.YAFFS2_ARCHIVE)

    # def test_implicit_type(self):
    #     self.usage_with(_p.join(self.test_dir.name, 'yaffs2_OOB.img'))

    def test_extract(self):
        self.usage_extract(_p.join(self.test_dir.name, 'yaffs2_OOB.img'), type=ArchiveType.YAFFS2_ARCHIVE)


class TestDiskImage(TestMountable):
    def __init__(self, *args, **kwargs):
        self.test_vectors = ['disk.img']
        super().__init__(*args, **kwargs)

    def test_with(self):
        file = _p.join(self.test_dir.name, 'disk.img')

        img = DiskImage(file)

        for i, partition in enumerate(img.partitions):
            with mount(partition) as m:
                with open(_p.join(m.name, 'hello'), 'r') as f:
                    assert f.read().strip() == 'hello' + str(i + 1)

    # def test_no_with(self):
    #     self.usage_no_with(_p.join(self.test_dir.name, 'ext4.img'), type=MountableType.LINUX_DISK_IMAGE)
    #
    # def test_implicit_type(self):
    #     self.usage_with(_p.join(self.test_dir.name, 'ext4.img'))



import re

import sh

from ..core import Mountable
from ..udisksctl import udisksctl, udisksctl_losetup, udisksctl_mount


class UDisksCtlCtx:
    __dev: str = None
    __mount_point: str = None

    def mount(self, image: str):
        self.__dev = udisksctl_losetup(image)

        self.__mount_point = udisksctl_mount(self.__dev)

    def umount(self):
        if self.__mount_point:
            udisksctl('unmount', b=self.__dev)
            self.__mount_point = None

        if self.__dev:
            udisksctl('loop-delete', b=self.__dev)
            self.__dev = None


class LinuxDiskImage(Mountable):
    __offset: int
    __lomode: str = 'mount'
    __mntmode: str = 'mount'
    __dev: str = None

    def __init__(self, file: str, offset=0, lomode='mount', mntmode='mount'):
        super().__init__(file)

        if not (lomode == 'mount' or lomode == 'udisksctl' or lomode == 'losetup'):
            raise ValueError

        # dhkim: TODO
        # if not (mntmode == 'mount' or mntmode == 'udisksctl'):
        if not (mntmode == 'mount'):
            raise ValueError

        self.__offset = offset
        self.__lomode = lomode
        self.__mntmode = mntmode

    def _mount(self, file: str, mode: str, mount_point: str):
        if self.__lomode == 'mount':
            sh.sudo.mount(file, mount_point, o=f'offset={self.__offset}', _fg=True)
            return
        elif self.__lomode == 'losetup':
            self.__dev = str(sh.losetup(f=True))
            sh.sudo.losetup(self.__dev, file, _fg=True)
        elif self.__lomode == 'udisksctl':
            self.__dev = udisksctl_losetup(file)
        else:
            raise ValueError

        print('mount: ' + self.__dev)

        assert self.__dev

        if self.__mntmode == 'mount':
            sh.sudo.mount(self.__dev, mount_point, o=f'loop,offset={self.__offset}', _fg=True)
        elif self.__mntmode == 'udisksctl':
            mount_point = udisksctl_mount(self.__dev)
        else:
            raise ValueError

    def _umount(self, mount_point):
        if self.__lomode == 'mount':
            sh.sudo.umount(mount_point, _fg=True)
            return

        if self.__mntmode == 'mount':
            # when mounting a loop device, mount command changes the loop device. hence, self.__dev is invalid at here
            # sh.sudo.umount(self.__dev, _fg=True)
            sh.sudo.umount(mount_point, _fg=True)
        elif self.__mntmode == 'udisksctl':
            udisksctl('unmount', b=self.__dev)
        else:
            raise ValueError

        if self.__lomode == 'losetup':
            raise NotImplementedError
        elif self.__lomode == 'udisksctl':
            udisksctl('loop-delete', b=self.__dev)

    @classmethod
    def _ismountable(cls, path: str = None, file_cmd_out: str = None) -> bool:
        if file_cmd_out is None:
            return False
        return (
            bool(re.search(r'^Linux[^,]*filesystem data', file_cmd_out))
            or bool(re.search(r'\bFAT\b', file_cmd_out))
        )



import getpass
import re
import sys

import keyring as keyring
import sh
from keyring.errors import KeyringLocked


class __UdisksctlAuth:
    SERVICE = 'udisksctl'

    stdout: str = ''

    def udisksctl_auth(self, char, stdin):
        sys.stdout.write(str(char))
        sys.stdout.flush()
        self.stdout += char

        sin = None

        if self.stdout.endswith('): '):
            if not (m := re.search(fr'\s*(?P<no>\d+).*\({getpass.getuser()}\)', self.stdout)):
                print(self.stdout)
                raise Exception(fr'cannot find user {getpass.getuser()} from the list')
            sin = m['no'] + '\n'
        if self.stdout.endswith('Password: '):
            try:
                pw = keyring.get_password(self.SERVICE, getpass.getuser())
                assert pw
            except KeyringLocked:
                pw = getpass.getpass(prompt='')
            except:
                pw = getpass.getpass(prompt='')
                keyring.set_password(self.SERVICE, getpass.getuser(), pw)
            stdin.put(pw + '\n')

            self.stdout = ''

        if sin:
            stdin.put(sin)
            sys.stdout.write(sin)
            sys.stdout.flush()


def udisksctl(*args, _auth: bool = False, **kwargs) -> str:
    if _auth:
        uauth = __UdisksctlAuth()

        sh.udisksctl(*args, **kwargs,
                     _tty_in=True, _tty_out=True, _unify_ttys=True,
                     _out=uauth.udisksctl_auth, _out_bufsize=0)
        return uauth.stdout
    else:
        return str(sh.udisksctl(*args, **kwargs))


def udisksctl_losetup(image) -> str:
    try:
        sout = udisksctl('loop-setup', file=image, _auth=True)
    except Exception:
        raise Exception(fr'failed to open {image}')

    if not (m := re.search(fr'Mapped file {image} as (?P<dev>[^\s]*loop[^\s]*)\.', sout)):
        print(sout)
        raise Exception(fr'failed to map a loop device for {image}')

    print(sout)

    return m['dev']


def udisksctl_mount(loop_dev) -> str:
    try:
        sout = udisksctl('mount', b=loop_dev, o='rw', _auth=True)
    except Exception:
        raise Exception('')

    if not (m := re.search(fr'Mounted {loop_dev} at (?P<dir>[^\s]*)\.', sout)):
        raise Exception(fr'failed to mount {loop_dev}')

    return m['dir']



from .cpio_archive import CpioArchive
from .disk_image import DiskImage
from .linux_disk_image import LinuxDiskImage
from .qcow2_image import Qcow2Image
from .qemu_qcow2_image import QemuQcow2Image
from .yaffs2_archive import Yaffs2Archive



# Ignore import errors
# https://stackoverflow.com/a/6077117/3836385

import builtins
from types import ModuleType

# argcomplete does not work when import errors are ignored.
try:
    import argcomplete
except ImportError:
    pass


class DummyModule(ModuleType):
    def __getattr__(self, key):
        return None

    __all__ = []  # support wildcard imports


def tryimport(name, globals={}, locals={}, fromlist=[], level=-1):
    try:
        return realimport(name, globals, locals, fromlist, level)
    except ImportError:
        return DummyModule(name)


realimport, builtins.__import__ = builtins.__import__, tryimport

from .mputil import MpUtil, globalize
from .subcommand import Subcommand, SubcommandParser
from .fileutil import remove_contents, find_recursive, overwrite, get_memtmpdir, find
from .decorators import static_vars
from .eventutil import EventGenerator, Event
from .strutil import decomment_cxx, strcompare
from .plotutil import plot_cdf, plot_linear, plot_scatter, plot_box, plot_stack
from .cacheutil import PersistentCache
from .dockerutil import docker_sh
from .gitutil import git_expand
from .pyshutil import compile_shargs

builtins.__import__ = realimport



import os
import re

# https://stackoverflow.com/a/241506/3836385
from itertools import zip_longest


def decomment_cxx(text):
    def replacer(match):
        s = match.group(0)
        if s.startswith('/'):
            return " " # note: a space and not an empty string
        else:
            return s
    pattern = re.compile(
        r'//.*?$|/\*.*?\*/|\'(?:\\.|[^\\\'])*\'|"(?:\\.|[^\\"])*"',
        re.DOTALL | re.MULTILINE
    )
    return re.sub(pattern, replacer, text)


def highlight_str(string: str, start: int, end: int):
    color = '\033[{0}m'
    color_str = color.format(31)  # red
    reset_str = color.format(0)

    hl_str = ''
    hl_str += string[0: start]
    hl_str += color_str
    hl_str += string[start: end]
    hl_str += reset_str
    hl_str += string[end:]

    return hl_str


def auto_print(string: str, **kwargs):
    if os.fstat(0) == os.fstat(1):
        print(string, **kwargs)
    else:
        print(re.sub(r'\033\[[0-9,;]*[m,K]', '', string), **kwargs)


def strcompare(left: str, right: str, width: int = None, highlight: bool = True) -> str:
    def __compose_line(ll: str, rl: str, padding: int):
        return str('{:' + str(len(str(ll)) + padding) + '} | {}').format(str(ll), str(rl)) + '\n'

    lls = left.splitlines()
    rls = right.splitlines()

    color_fmt = '\033[{0}m'
    color_red = color_fmt.format(31)  # red
    color_clr = color_fmt.format(0)

    if not width:
        width = max([len(e) for e in lls])

    diff = ""
    for ll, rl in zip_longest(lls, rls):
        ll: str
        rl: str

        padding = width - len(str(ll))

        if ll == rl:
            diff += __compose_line(ll, rl, padding)
            continue

        if highlight and ll and rl:
            hl = False
            i = 0
            while i < min(len(ll), len(rl)):
                if (not hl) and (ll[i] != rl[i]):
                    hl = True
                    ll = ll[:i] + color_red + ll[i:]
                    rl = rl[:i] + color_red + rl[i:]
                    i += len(color_red) + 1
                elif (hl) and (ll[i] == rl[i]):
                    hl = False
                    ll = ll[:i] + color_clr + ll[i:]
                    rl = rl[:i] + color_clr + rl[i:]
                    i += len(color_clr) + 1
                else:
                    i += 1

            if hl:
                ll += color_clr
                rl += color_clr

        diff += __compose_line(ll, rl, padding)
    return diff




import os
import re
import shutil
from tempfile import TemporaryDirectory
from typing import Union
from os import path as _p

import sh


class AlreadyMountedError(Exception):
    pass


class NotMountedError(Exception):
    pass


def _empty_dir(dir: str):
    for filename in os.listdir(dir):
        file_path = os.path.join(dir, filename)
        try:
            if os.path.isfile(file_path) or os.path.islink(file_path):
                os.unlink(file_path)
            elif os.path.isdir(file_path):
                shutil.rmtree(file_path)
        except Exception as e:
            print('Failed to delete %s. Reason: %s' % (file_path, e))


class MountPoint:
    __mountable: None
    __mode: str
    __mount_point: str
    __tmpdir: TemporaryDirectory

    def __init__(self, mountable, mode: str, mount_point: str):
        self.__mountable = mountable
        self.__mode = mode

        if not mount_point:
            self.__tmpdir = TemporaryDirectory(prefix='yautil-mountutil-')
            mount_point = self.__tmpdir.name

        self.__mount_point = mount_point

    @property
    def name(self) -> str:
        return self.__mount_point

    def umount(self):
        if not self.__mountable._is_mounted:
            raise NotMountedError('Not mounted')

        self.__mountable._is_mounted = False
        return self.__mountable._umount(self.name)

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        try:
            self.umount()
        except NotMountedError:
            pass

    def __del__(self):
        try:
            self.umount()
        except NotMountedError:
            pass


class Mountable:
    __file: str
    _is_mounted: bool = False

    def __init__(self, file: str, **kwargs):
        if type(self) == Mountable:
            if not _p.isfile(file):
                raise FileNotFoundError(file)

            if not (cls := self.__resolve_cls(file)):
                raise Exception('unsupported file type ' + file + ' ' + str(sh.file(file)))
            self.__class__ = cls
            self.__init__(file, **kwargs)

        assert type(self) != Mountable
        assert isinstance(self, Mountable)

        self.__file = file

    def _mount(self, file: str, mode: str, mount_point: str):
        raise NotImplementedError

    def _umount(self, mount_point):
        raise NotImplementedError

    @classmethod
    def _ismountable(cls, path: str = None, file_cmd_out: str = None) -> bool:
        raise NotImplementedError

    @property
    def partitions(self) -> Union[list, None]:
        return None

    @property
    def name(self) -> str:
        return self.__file

    @classmethod
    def __resolve_cls(cls, file: str):
        from .types import MountableType

        desc = str(sh.file(file, brief=True)).strip()

        for typ in MountableType:
            if typ == MountableType.AUTO:
                continue
            mountable_cls = typ.value
            if mountable_cls._ismountable(file, desc):
                return mountable_cls


class Archive(Mountable):
    ro: bool = False

    def _extract(self, file: str, target_dir: str):
        raise NotImplementedError

    def _archive(self, file: str, source_dir: str):
        raise NotImplementedError

    def _mount(self, file: str, mode: str, mount_point: str):
        if mode == 'r':
            self.ro = True
        elif mode == 'rw':
            self.ro = False
        else:
            raise ValueError('mode must be either \'r\' or \'rw\'')

        self._extract(file, mount_point)

    def _umount(self, mount_point):
        if not self.ro:
            self._archive(self.name, mount_point)

        _empty_dir(mount_point)



import sh

from yautil.mountutil import Archive


class CpioArchive(Archive):
    def _extract(self, file: str, target_dir: str):
        with open(file, 'r') as f:
            sh.cpio(sh.gzip(d=True, k=True, _in=f), i=True, _cwd=target_dir)

    def _archive(self, file: str, source_dir: str):
        # sh.gzip(sh.mkbootfs(source_dir), _out=file)
        raise NotImplementedError

    @classmethod
    def _ismountable(cls, path: str = None, file_cmd_out: str = None) -> bool:
        if 'gzip compressed data' in file_cmd_out:
            file_cmd_out = sh.file(path, z=True)

        return 'ASCII cpio archive' in file_cmd_out



from enum import Enum

from .core import Mountable, Archive
from .mountables import LinuxDiskImage, QemuQcow2Image, Yaffs2Archive, CpioArchive, Qcow2Image


class MountableType(Enum):
    AUTO = Mountable
    LINUX_DISK_IMAGE = LinuxDiskImage
    QEMU_QCOW2_IMAGE = QemuQcow2Image
    QCOW2_IMAGE = Qcow2Image
    CPIO_ARCHIVE = CpioArchive
    YAFFS2_ARCHIVE = Yaffs2Archive

    def instantiate(self, file: str, **type_specific_args) -> Mountable:
        return self.value(file, **type_specific_args)


class ArchiveType(Enum):
    AUTO = Archive
    CPIO_ARCHIVE = CpioArchive
    YAFFS2_ARCHIVE = Yaffs2Archive

    def instantiate(self, file: str) -> Archive:
        return self.value(file)



import os
from os import path as _p
from shutil import copyfile
import sh


def link_repo(base: str, target: str):
    if not _p.isdir(base):
        raise NotADirectoryError(base)

    if not _p.isdir(target):
        raise NotADirectoryError(target)

    base_repo = _p.realpath(_p.join(base, '.git'))

    if not _p.isdir(base_repo):
        raise NotADirectoryError(base_repo)

    target_repo = _p.realpath(_p.join(target, '.git'))

    if _p.exists(target_repo):
        raise FileExistsError(target_repo)

    os.mkdir(target_repo)

    items_to_copy = {'HEAD', 'ORIG_HEAD'}
    items_to_omit = {'index'}
    items_to_step_in = {'logs'}
    items_to_ignore = []

    for root, dirs, files in os.walk(base_repo):
        root_path = _p.relpath(root, base_repo)
        skip = [*filter(lambda s: root_path == s or root_path.startswith(s + _p.sep), items_to_ignore)]
        if skip:
            continue

        for item in dirs + files:
            item_path = _p.relpath(_p.join(root, item), base_repo)
            if item in items_to_copy:
                assert _p.isfile(_p.join(base_repo, item_path))
                copyfile(_p.join(base_repo, item_path), _p.join(target_repo, item_path))
            elif item in items_to_omit:
                pass
            elif item in items_to_step_in:
                assert _p.isdir(_p.join(base_repo, item_path))
                os.mkdir(_p.join(target_repo, item_path))
            else:
                os.symlink(_p.join(base_repo, item_path), _p.join(target_repo, item_path))
                items_to_ignore.append(item_path)


def __git_expand(repo: str, dest: str, *checkout_targets: str, ignore_errors=False):
    if not _p.isdir(repo):
        raise NotADirectoryError(repo)

    if not _p.isdir(dest):
        raise NotADirectoryError(dest)

    dest_items = os.listdir(dest)

    if dest_items:
        for checkout_target in checkout_targets:
            if checkout_target in dest_items:
                raise FileExistsError(_p.join(dest, checkout_target))

    for checkout_target in checkout_targets:
        target_repo = _p.realpath(_p.join(dest, checkout_target))
        try:
            os.makedirs(target_repo, exist_ok=True)
            link_repo(repo, target_repo)
            sh.git.checkout(checkout_target, _cwd=target_repo)
        except:
            if ignore_errors:
                yield None
            else:
                raise Exception(f'Failed to checkout {checkout_target}')

        yield target_repo


def git_expand(repo: str, dest: str, *checkout_targets: str, ignore_errors=False, iter=False):
    if iter:
        return __git_expand(repo, dest, *checkout_targets, ignore_errors=ignore_errors)
    else:
        return [*__git_expand(repo, dest, *checkout_targets, ignore_errors=ignore_errors)]



import sh
from os import path as _p

from yautil.mountutil import Archive


YAFFS2_SRC_DIR = _p.join(_p.dirname(__file__), _p.pardir, 'yaffs2utils-0.2.9')


class Yaffs2Archive(Archive):
    mkyaffs2: callable = None
    unyaffs2: callable = None

    def _extract(self, file: str, target_dir: str):
        if not self.unyaffs2:
            sh.make(_cwd=YAFFS2_SRC_DIR)
            self.unyaffs2 = sh.Command('unyaffs2', [YAFFS2_SRC_DIR])

        # --yaffs-ecclayout: https://github.com/djeclipser/yaffs2utils/issues/35#issuecomment-153830017
        self.unyaffs2(file, target_dir, yaffs_ecclayout=True)

    def _archive(self, file: str, source_dir: str):
        if not self.mkyaffs2:
            sh.make(_cwd=YAFFS2_SRC_DIR)
            self.mkyaffs2 = sh.Command('mkyaffs2', [YAFFS2_SRC_DIR])

        self.mkyaffs2(source_dir, file, yaffs_ecclayout=True)

    @classmethod
    def _ismountable(cls, path: str = None, file_cmd_out: str = None) -> bool:
        return r'VMS Alpha Executable' in file_cmd_out



import glob
import re
import time
from typing import Union, List
import os
from os import path as _p

import sh

from yautil.mountutil.core import Mountable
from yautil.mountutil.block_dev_handle import BlockDevHandle


class QemuNbdHandle(BlockDevHandle):
    __img: str
    __dev: str = None
    __load_nbd = False

    def __init__(self, img: str):
        self.__img = img

    def find_free_nbd_dev(self):
        nbd_dev_avail = None
        modules, sizes, deps = zip(*[l.split(maxsplit=2) for l in str(sh.lsmod()).splitlines()])

        if 'nbd' not in modules:
            sh.sudo.modprobe('nbd', 'max_part=8', _fg=True)
            self.__load_nbd = True

        if not (nbd_devs := [*filter(lambda d: re.search(r'^nbd\d+$', d), os.listdir('/dev'))]):
            self.unmap()
            raise Exception('Failed to load nbd')

        for nbd_dev in nbd_devs:
            with open(fr'/sys/class/block/{nbd_dev}/size') as f:
                if f.read() == '0\n':
                    nbd_dev_avail = nbd_dev
                    break

        if not nbd_dev_avail:
            self.unmap()
            raise Exception('No nbd is available for mounting a qemu image')

        return f'/dev/{nbd_dev_avail}'

    def map(self) -> str:
        if self.base_dev:
            return self.base_dev

        dev = self.find_free_nbd_dev()

        assert dev

        try:
            sh.sudo('qemu-nbd', '--connect', dev, '-f', 'qcow2', self.__img, _fg=True)
            time.sleep(1)
        except sh.ErrorReturnCode_1:
            raise Exception(f'failed to open {self.__img}')

        self.__dev = dev

        return self.base_dev

    def unmap(self):
        if self.__dev:
            sh.sudo('qemu-nbd', '--disconnect', self.__dev, _fg=True)
            self.__dev = None

    @property
    def base_dev(self) -> str:
        return self.__dev

    @property
    def devs(self) -> List[str]:
        return [*filter(lambda d: re.search(fr'{self.base_dev}[a-zA-Z]', d), glob.glob('/dev/*'))]

    def __del__(self):
        self.unmap()


class QemuQcow2Image(Mountable):
    __nbd_ctx_: QemuNbdHandle = None
    __dev: str = None
    __partitions: List[Mountable] = None

    def __init__(self, file: str, dev: str = None):
        super().__init__(file)
        self.__dev = dev

    def _mount(self, file: str, mode: str, mount_point: str):
        assert self.__dev

        sh.sudo.mount(self.__dev, mount_point, _fg=True)

    def _umount(self, mount_point):
        sh.sudo.umount(mount_point, _fg=True)

    @classmethod
    def _ismountable(cls, path: str = None, file_cmd_out: str = None) -> bool:
        if file_cmd_out is None:
            return False
        return bool(re.search(r'^QEMU QCOW2 Image', file_cmd_out))

    @property
    def partitions(self) -> Union[list, None]:
        if self.__dev:
            return None

        if self.__partitions:
            return self.__partitions

        self.__nbd_ctx.map()
        self.__partitions = [QemuQcow2Image(self.name, dev=dev) for dev in self.__nbd_ctx.devs]

        return self.__partitions

    @property
    def __nbd_ctx(self) -> QemuNbdHandle:
        if not self.__nbd_ctx_:
            self.__nbd_ctx_ = QemuNbdHandle(self.name)
        return self.__nbd_ctx_



from typing import Union
from os import path as _p

from .core import MountPoint, Mountable
from .types import MountableType, ArchiveType


def mount(file: Union[str, Mountable], mode: str = 'rw', mount_point: str = None, type: MountableType = MountableType.AUTO,
          **type_specific_args) -> MountPoint:

    if not isinstance(file, Mountable):
        if not _p.isfile(file):
            raise FileNotFoundError(file)

        file = type.instantiate(file, **type_specific_args)

    # if mountable._is_mounted:
    #     raise AlreadyMountedError('Already mounted')

    if file.partitions:
        raise Exception('Specify one of partitions.')

    mp = MountPoint(file, mode, mount_point)

    file._is_mounted = True
    file._mount(file.name, mode, mp.name)

    return mp


def extract(file: str, dest: str, type: ArchiveType = ArchiveType.AUTO):
    if not _p.isfile(file):
        raise FileNotFoundError(file)

    a = type.instantiate(file)

    a._extract(file, dest)


def archive(file: str, src: str, type: ArchiveType):
    if type == ArchiveType.AUTO:
        raise ValueError(type)

    a = type.instantiate(file)

    a._archive(file, src)



import time

import sh as sh

from ..core import Mountable


class Qcow2Image(Mountable):
    def _mount(self, file: str, mode: str, mount_point: str):
        sh.nbdfuse(mount_point, '--socket-activation', 'qemu-nbd', file, _bg=True)
        time.sleep(1)

    def _umount(self, mount_point):
        # use '-uz' to force unmount https://stackoverflow.com/a/25986155/3836385
        sh.fusermount3('-uz', mount_point)

    @classmethod
    def _ismountable(cls, path: str = None, file_cmd_out: str = None) -> bool:
        if file_cmd_out is None:
            return False
        return bool(r'QCOW2' in file_cmd_out)



import re
from typing import Union, List

import sh

from ..core import Mountable
from .linux_disk_image import LinuxDiskImage


class DiskImage(Mountable):
    __partitions: List[Mountable] = None
    __in_qcow2: bool = False

    def __init__(self, file: str, in_qcow2: bool = False):
        super().__init__(file)
        self.__in_qcow2 = in_qcow2

    def __iter_partitions(self):
        rc = sh.fdisk(self.name, l=True, o='Start,End,Sectors,Type', color='never', _iter=True)
        sector_size = 512

        for line in rc:
            line = str(line).strip()

            if m := re.search(r'Units:.*?(?P<sector_size>\d+) bytes', line):
                sector_size = int(m['sector_size'])

            if not line:
                break

        for line in rc:
            line = str(line).strip()

            try:
                # be aware that 'type' may include spaces.
                start, end, sectors, type = line.split(maxsplit=3)
                start, end, sectors = (int(i) for i in [start, end, sectors])
            except ValueError:
                # skip column titles
                continue

            assert end - start + 1 == sectors
            # print(f'start: {start}, sector_size: {sector_size}')

            yield start * sector_size, sectors * sector_size, type.strip()

    def _mount(self, file: str, mode: str, mount_point: str):
        assert False

    def _umount(self, mount_point):
        assert False

    @classmethod
    def _ismountable(cls, path: str = None, file_cmd_out: str = None) -> bool:
        if file_cmd_out is None:
            return False
        return r'boot sector' in file_cmd_out

    @property
    def partitions(self) -> Union[List[Mountable], None]:
        if self.__partitions:
            return self.__partitions

        self.__partitions = []
        for start, size, type in self.__iter_partitions():
            if not type.startswith('Linux'):
                raise NotImplementedError('only Linux filesystem partitions are supported for now')
            self.__partitions.append(
                LinuxDiskImage(self.name, offset=start, lomode='udisksctl' if self.__in_qcow2 else 'mount')
            )

        return self.__partitions



import getpass
import os
import sys
import tempfile
from os import path as _p
from typing import Union, List
# from deprecation import deprecated

import sh


def __build(build_context, fg=False, dockerfile_cmds_to_append: list = None, drop_priv=False, kvm=False):
    with open(_p.join(build_context, 'Dockerfile'), 'r') as f:
        dockerfile = f.read()

    if dockerfile_cmds_to_append:
        dockerfile += '\n{cmds}\n'.format(cmds='\n'.join(dockerfile_cmds_to_append))

    if kvm:
        import grp
        kvm_gid = grp.getgrnam('kvm').gr_gid
        dockerfile += (
            # f'RUN apt-get install -y qemu-kvm libvirt-bin ubuntu-vm-builder bridge-utils'
            # f'\n'
            # f'RUN groupmod -g {kvm_gid} kvm'
            f'\n'
            f'RUN groupadd -g {kvm_gid} kvm'
            f'\n'
        )

    # if xforwarding:
    #     dockerfile += (
    #         f'\n'
    #         f'RUN apt-get install -y xinit xserver-xorg-core --no-install-recommends --no-install-suggests'
    #         f'\n'
    #     )

    if drop_priv:
        username = getpass.getuser()
        uid = os.getuid()
        gid = os.getgid()
        home = f'/home/{username}'
        groupname = f'$(getent group {gid} | cut -d: -f1)'

        dockerfile += (
            f'\n'
            f'SHELL ["/bin/bash", "-c"]'
            f'\n'
            f'RUN if [ "$(id -u {username} > /dev/null 2>&1; echo $?)" == 0 ]; then userdel {username}; fi'
            f' && groupadd -g {gid} {username}'
            f' ;  useradd -l -u {uid} -g {groupname} {"-G kvm" if kvm else ""} {username}'
            f' && install -d -m 0755 -o {username} -g {groupname} {home}'
            f'\n'
            f'SHELL ["/bin/sh", "-c"]'
            f'\n'
            f'ENV HOME={home}'
            f'\n'
            f'ENV USER={username}'
            f'\n'
            f'USER {uid}:{gid}'
            f'\n'
        )

    tmpdir = tempfile.TemporaryDirectory()

    iidfile = _p.join(tmpdir.name, '__iid')
    try:
        sh.docker.build('.',
                        f='-',
                        iidfile=iidfile,
                        _in=dockerfile,
                        _cwd=build_context,
                        _err_to_out=bool(fg),
                        _out=sys.stdout if bool(fg) else None,
                        )
    except sh.ErrorReturnCode as e:
        raise Exception(f'Failed to build a docker image with build context at {build_context}.\n'
                        f'STDOUT:\n'
                        f'{bytes(e.stdout).decode(sh.DEFAULT_ENCODING)}'
                        f'\n'
                        f'STDERR:\n'
                        f'{bytes(e.stderr).decode(sh.DEFAULT_ENCODING)}'
                        )
    except Exception:
        raise Exception(f'failed to build a docker image with build context at {build_context}.')

    with open(iidfile, 'r') as f:
        return f.read()


def docker_sh(
        docker_context: str,
        *docker_run_opts,
        root: bool = False,
        verbose: bool = False,
        volumes: Union[str, List[str]] = None,
        auto_remove: bool = True,
        dockerfile_cmds_to_append: list = None,
        kvm: bool = False,
        xforwarding: bool = False,
        net: str = 'bridge',
        _cwd: str = None,
) -> sh.Command:

    if (not docker_context) or (not _p.isdir(docker_context)):
        raise Exception('proper docker_context directory must be supplied')

    if os.getuid() == 0:
        root = True

    if verbose:
        print('Building a docker image...')
    image_id = __build(docker_context, fg=verbose, drop_priv=not root, kvm=kvm,
                       dockerfile_cmds_to_append=dockerfile_cmds_to_append)
    if not image_id:
        raise Exception('failed to build image')

    if root:
        home = '/root'
    else:
        username = getpass.getuser()
        home = f'/home/{username}'

    if docker_run_opts is None:
        docker_run_opts = []
    else:
        docker_run_opts = list(docker_run_opts)

    if not volumes:
        pass
    elif isinstance(volumes, str):
        docker_run_opts += [f'-v={volumes}']
    elif isinstance(volumes, list):
        docker_run_opts += [*map(lambda o: f'-v={o}', volumes)]
    else:
        raise Exception

    if xforwarding:
        # run_opts.append('-v=/tmp/.X11-unix:/tmp/.X11-unix:rw')
        docker_run_opts.append(f'-v={_p.join(os.environ["HOME"], ".Xauthority")}:{_p.join(home, ".Xauthority")}')
        docker_run_opts.append(f'-eDISPLAY={os.environ["DISPLAY"]}')
        net = 'host'

    if kvm:
        docker_run_opts.append('--device=/dev/kvm')
        docker_run_opts.append('--group-add=kvm')
        # run_opts.append(f'-v=/etc/machine-id:/etc/machine-id:rw')
        # run_opts.append(f'-eQEMU_AUDIO_DRV=none')

    docker_run_opts.append(f'--net={net}')

    run = sh.docker.run.bake(
        *docker_run_opts,
        '-d=false',
        i=True,
        rm=bool(auto_remove),  # Automatically remove the container when it exits
        workdir=_p.realpath(_cwd) if _cwd else home,
    )

    return run.bake(image_id)



import setuptools
import os

# borrowed from https://stackoverflow.com/a/36693250
def package_files(directory):
    paths = []
    for (path, directories, filenames) in os.walk(directory):
        for filename in filenames:
            paths.append(os.path.join('..', path, filename))
    return paths


deps = [
    'sh',
    'matplotlib',
]

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="yautil", # Replace with your own username
    version="0.0.61",
    author="Donghwi Kim",
    author_email="dhkim09@kaist.ac.kr",
    description="Yet Another Python util.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/dhkim09a/pyutil",
    packages=setuptools.find_packages(),
    # classifiers=[
    #     "Programming Language :: Python :: 3",
    #     "License :: OSI Approved :: MIT License",
    #     "Operating System :: OS Independent",
    # ],
    install_requires=deps,
    extras_require={
        'test': deps,
    },
    package_data={'': package_files('yautil/mountutil/yaffs2utils-0.2.9')},
    python_requires='>=3.7',
)
